/*
 * irc.c
 */

__asm("	text\n	dc.b	'$Id: irc.c,v 1.21 1998/09/07 21:09:11 sharl Exp $'\n	even\n");

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#define __IOCS_INLINE__
#include <iocslib.h>

#include "config.h"
#include "console.h"
#include "ircdef.h"
#include "irc.h"
#include "action.h"
#include "command.h"
#include "channel.h"
#include "util.h"
#include "kanji.h"

/****************************************
	T[õbZ[W
****************************************/
static BOOLEAN
irc_fromserver(int sock, char *buf)
{
	char newbuf[1024], tmpbuf[1024];
	char chan[512], chanbuf[512];
	int num, len;

	if (sscanf(buf, "%*s %d %*s %n", &num, &len)) {
#ifdef DEBUG
		if (debug)
			printf(buf);
#endif
		switch (num){
			case 1: case 2: case 3:
			case RPL_LUSERCLIENT:
			case RPL_LUSERME:
			case RPL_MOTDSTART:
			case RPL_UNAWAY:
			case RPL_NOWAWAY:
				printf("%s \033[31m%s\033[m", getnow(), buf + len + 1);
				break;
			case RPL_INFO:
			case RPL_MOTD:
				jstr2sjstr(newbuf, buf + len + 1);
				printf("%s \033[31m%s\033[m", getnow(), newbuf);
				break;
			case 4:
				printf("%s \033[31m%s\033[m", getnow(), buf + len);
				break;
			case RPL_LUSERUNKNOWN:
			case RPL_LUSERCHANNELS:
				{
					int number, n;

					sscanf(buf + len, "%d :%n", &number, &n);
					printf("%s \033[31m%d %s\033[m", getnow(), number, buf + len + n);
				}
				break;
			case RPL_INVITING:
				jstr2sjstr(newbuf, buf + len);
				printf("%s \033[31mINVITE %s\033[m", getnow(), newbuf);
				break;
			case RPL_NAMREPLY:
				auto_oper(sock, buf + len);
				break;
			case RPL_WHOREPLY:
			case RPL_ENDOFWHO:
				jstr2sjstr(newbuf, buf + len);
				printf("%s \033[31m*** %s\033[m", getnow(), newbuf);
				break;
			case RPL_ISON:
				if (ison)
					notify(buf + len);
				else
					printf("%s \033[31mISON %s\033[m", getnow(), buf + len);
				break;
			case RPL_WHOISUSER:
			case RPL_WHOWASUSER:
				{
					int n;

					sscanf(buf + len, "%s %n", tmpbuf, &n);
					jstr2sjstr(newbuf, buf + len + n);
					printf("%s \033[31m*** %s: %s\033[m", getnow(), tmpbuf, newbuf);
				}
				break;
			case RPL_WHOISSERVER:
				{
					char comm[256];

					sscanf(buf + len, "%s %s :%[^\r\n]", tmpbuf, newbuf, comm);
					printf("%s \033[31m*** %s: %s [%s]\033[m\n", getnow(), tmpbuf, newbuf, comm);
				}
				break;
			case RPL_WHOISIDLE:
				{
					int t, n;

					sscanf(buf + len, "%s %d :%n", tmpbuf, &t, &n);
					printf("%s \033[31m*** %s: %d %s\033[m", getnow(), tmpbuf, t, buf + len + n);
				}
				break;
			case RPL_LIST:
				{
					int t, n;

					sscanf(buf + len, "%s %d :%n", chan, &t, &n);
					jchan2sjchan(chanbuf, chan);
					jstr2sjstr(tmpbuf, buf + len + n);
					printf("%s \033[31m*** %s: %d: %s\033[m", getnow(), chanbuf, t, tmpbuf);
				}
				break;
			case RPL_LISTEND:
				printf("%s \033[31m*** %s\033[m", getnow(), buf + len + 1);
				break;
			case RPL_AWAY:
			case RPL_ENDOFWHOIS:
			case RPL_WHOISCHANNELS:
			case RPL_ENDOFWHOWAS:
			case RPL_ENDOFNAME:
				{
					int n;

					sscanf(buf + len, "%s :%n", chan, &n);
					jchan2sjchan(chanbuf, chan);
					jstr2sjstr(tmpbuf, buf + len + n);
					printf("%s \033[31m*** %s: %s\033[m", getnow(), chanbuf, tmpbuf);
				}
				break;
			case RPL_CHANNELMODEIS:
				jstr2sjstr(newbuf, buf + len);
				printf("%s \033[31mMODE %s\033[m", getnow(), newbuf);
				break;
			case RPL_TOPIC:
				sscanf(buf + len, "%s :%[^\r\n]", chan, tmpbuf);
				jchan2sjchan(chanbuf, chan);
				jstr2sjstr(newbuf, tmpbuf);
				printf("%s \033[31mTOPIC %s %s\033[m\n", getnow(), chanbuf, newbuf);
				set_topic(chanbuf, newbuf);
				break;
			case RPL_VERSION:
				printf("%s \033[31mVERSION %s\033[m", getnow(), buf + len);
				break;
			case RPL_TIME:
				sscanf(buf + len, "%s :%[^\r\n]", newbuf, tmpbuf);
				printf("%s \033[31m%s: %s\033[m\n", getnow(), newbuf, tmpbuf);
				break;
			case RPL_ENDOFINFO:
			case RPL_ENDOFMOTD:
			case ERR_NOTEXTTOSEND:
				break;
			case ERR_USERSDONTMATCH:
				printf("%s \033[36m%s\033[m", getnow(), buf + len + 1);
				break;
			case ERR_NICKNAMEINUSE:
				printf("%s \033[32m%s\033[m", getnow(), buf + len);
				return FALSE;
			case ERR_UNKNOWNCOMMAND:
				jstr2sjstr(newbuf, buf + len);
				printf("\033[36m%s\033[m", newbuf);
				break;
			default:
				jstr2sjstr(newbuf, buf + len);
				printf("%s \033[36m%d\033[m %s", getnow(), num, newbuf);
				break;
		}
	} else {
		printf("%s \033[36m%s\033[m", getnow(), buf);
	}
	return TRUE;
}

/****************************************
	[ŨbZ[W
****************************************/
static void
irc_fromuser(char *buf)
{
	char newbuf[1024];
	char from[512];
	char comm[512];
	char to[512];
	int len;

	sscanf(buf, "%s %s %s %n", from, comm, to, &len);
	if (!strcmp(from, irc_nick) && !strcmp(to, irc_nick) && !strcmp(comm, "MODE")) {
		printf("%s Your MODE is set %s", getnow(), buf + len + 1);
	} else {
		jstr2sjstr(newbuf, buf);
		printf("%s \033[36m%s\033[m", getnow(), newbuf);
	}
}

/****************************************
	CTCP 
****************************************/
static void
irc_ctcpfrom(int sock, char *nick, char *buf)
{
	char line[1024];
	char newbuf[1024];

	if (!strncasecmp(buf, "\001PING", 5)) {
		time_t ping;

		if (sscanf(buf + 1, "%*s %d\001", &ping) != EOF) {
			printf("%s \033[31mCTCP PING from %s\033[m\n", getnow(), nick);
			sprintf(line, "NOTICE %s :\001PING %d\001\r\n", nick, ping);
#ifdef DEBUG
			if (debug)
				printf(line);
#endif
			write(sock, line, strlen(line));
		}
	} else if (!strcasecmp(buf, "\001TIME\001")) {
		time_t tim;
		struct tm *tm;
		char timbuf[256];

		tim = time(NULL);
		tm = localtime(&tim);
		printf("%s \033[31mCTCP TIME from %s\033[m\n", getnow(), nick);
		strftime(timbuf, sizeof(timbuf), "%a %b %d %T %Y", tm);
		sprintf(line, "NOTICE %s :\001TIME %s\001\r\n", nick, timbuf);
#ifdef DEBUG
		if (debug)
			printf(line);
#endif
		write(sock, line, strlen(line));
	} else if (!strcasecmp(buf, "\001VERSION\001")) {
		printf("%s \033[31mCTCP VERSION from %s\033[m\n", getnow(), nick);
		sprintf(line,
				"NOTICE %s :\001VERSION " PROGNAME " " VERSION " (X68k; " __DATE__ ")\001\r\n",
				nick);
		sjstr2jstr(newbuf, line);
#ifdef DEBUG
		if (debug)
			printf(line);
#endif
		write(sock, newbuf, strlen(newbuf));
	} else if (!strcasecmp(buf, "\001USERINFO\001")) {
		printf("%s \033[31mCTCP USERINFO from %s\033[m\n", getnow(), nick);
		sprintf(line, "NOTICE %s :\001USERINFO :%s\001\r\n", nick, irc_info);
		sjstr2jstr(newbuf, line);
#ifdef DEBUG
		if (debug)
			printf(line);
#endif
		write(sock, newbuf, strlen(newbuf));
	} else if (!strcmp(buf, "\001CLIENTINFO\001")) {
		printf("%s \033[31mCTCP CLIENTINFO from %s\033[m\n", getnow(), nick);
		sprintf(line,
				"NOTICE %s :\001CLIENTINFO :VERSION CLIENTINFO USERINFO PING TIME\001\r\n",
				nick);
#ifdef DEBUG
		if (debug)
			printf(line);
#endif
		write(sock, line, strlen(line));
	}
}

/****************************************
	s
****************************************/
BOOLEAN
irc_parseline(int sock, char *buf)
{
	const CMD *cmd;
	char newbuf[1024];
	char chanbuf[1024];
	char nick[512] = {0};
	char chan[512] = {0};
	char comm[512] = {0};
	char rest[512] = {0};

	if (*buf == ':') {
//:irc.hauN.org 001 Sharl :Welcome to the Internet Relay Network Sharl
// ꂪƎv
//:violet 001 Sharl :Welcome to the Internet Relay Network Sharl
//  response Ԃ̂ (THANX onda 1998/08/22)
		sscanf(buf + 1, "%s", newbuf);
		if (!strncmp(irc_server, newbuf, strlen(newbuf))
			|| !strncmp(newbuf, irc_server, strlen(irc_server))
		   ) {	/* T[õbZ[W */
			return irc_fromserver(sock, buf + 1);
		} else if (strchr(buf, '!') == NULL) {	/* [ŨbZ[W */
			irc_fromuser(buf + 1);
		} else {	/* ̑̃[ŨbZ[W */
			int len;

#ifdef DEBUG
			if (debug)
				printf(buf);
#endif
			sscanf(buf + 1, "%[^!]%*s %s %n", nick, comm, &len);

			buf += len + 1;
			if (*buf == ':') {	/* :hoho!~hoho@10.0.0.1 JOIN :#scot */
				char *p = chan;

				buf++;
				while (*buf && *buf != '\r' && *buf != '\a')
					*p++ = *buf++;
				*p = '\0';
			} else {			/* :hoho!~hoho@10.0.0.1 PRIVMSG #scot :hoe- */
				char *p = rest;
				int len;

				sscanf(buf, " %s %n", chan, &len);
				buf += len;
				while (*buf && *buf != '\r')
					*p++ = *buf++;
				*p = '\0';
			}

			jchan2sjchan(chanbuf, chan);
#ifdef DEBUG
			if (debug) {
				printf("nick [%s]\n", nick);
				printf("comm [%s]\n", comm);
				printf("chan [%s]\n", chanbuf);
				printf("rest [%s]\n", rest);
			}
#endif
			cmd = in_word_set(comm, strlen(comm));
			switch (cmd->cmd) {
				case T_INVITE:
					jchan2sjchan(newbuf, rest + 1);
					printf("%s \033[31m%s invites you to %s\033[m\n", getnow(), nick, newbuf);
					if (irc_autojoin) {
						char line[1024];

						printf("%s \033[32mauto join to %s\033[m\n", getnow(), newbuf);
						sprintf(line, "JOIN %s\r\n", rest + 1);
						write(sock, line, strlen(line));
					}
					break;
				case T_JOIN:
					printf("%s \033[35m%s has joined %s\033[m\n", getnow(), nick, chanbuf);
					if (!strcmp(irc_nick, nick)) {
						join_chanlist(chanbuf);
					} else {
						char line[1024];

						sprintf(line, "NAMES %s\r\n", chan);
						write(sock, line, strlen(line));
					}
					break;
				case T_KICK:
					jstr2sjstr(newbuf, rest);
					printf("%s \033[31m%s kicked %s out %s\033[m\n",
						   getnow(), nick, chanbuf, newbuf);
					if (!strncmp(irc_nick, rest, strlen(irc_nick)))
						part_chanlist(chanbuf);
					if (irc_autorejoin && !strncmp(newbuf, irc_nick, strlen(irc_nick))) {
						char linebuf[1024];

						sprintf(linebuf, "JOIN %s\r\n", chan);
						write(sock, linebuf, strlen(linebuf));
					}
					break;
				case T_MODE:
					printf("%s \033[31mNew mode %s %sby %s\033[m\n", getnow(), chanbuf, rest, nick);
					parse_mode(sock, chanbuf, rest);
					break;
				case T_NICK:
					printf("%s \033[31m%s is now %s\033[m\n", getnow(), nick, chan);
					break;
				case T_PART:
					printf("%s \033[35m%s has parted %s\033[m\n", getnow(), nick, chanbuf);
					if (!strcmp(irc_nick, nick))
						part_chanlist(chanbuf);
					break;
				case T_QUIT:
					printf("%s \033[35m%s has left IRC (%s)\033[m\n", getnow(), nick, chanbuf);
					break;
				case T_TOPIC:
					jstr2sjstr(newbuf, rest + 1);
					printf("%s \033[31mNew topic on %s by %s: %s\033[m\n", getnow(), chanbuf, nick, newbuf);
					set_topic(chanbuf, newbuf);
					break;

				case T_NOTICE:
					if (!strcasecmp(irc_nick, chan)) {
						int len;

						sscanf(rest, ":\001%s %n", comm, &len);
						if (!strcasecmp(comm, "PING")) {
							time_t ping;

							sscanf(rest + len, " %d", &ping);
							if (ping == ping_code) {
								printf("%s \033[31m%s(%s) = %d second\033[m\n",
									   getnow(), comm, nick, time(NULL) - irc_ping);
							} else {
								printf("\033[32mIllegal PING code returns from %s(%d)\033[m\n",
									   nick, ping);
							}
						} else if (!strcasecmp(comm, "TIME")
								   || !strcasecmp(comm, "VERSION")) {
							char *p = newbuf;
							char *q = rest + len;
							char tmpbuf[1024];

							while (*q && *q != '\001')
								*p++ = *q++;
							*p = '\0';
							jstr2sjstr(tmpbuf, newbuf);
							printf("%s \033[31m%s(%s) = %s\033[m\n", getnow(), comm, nick, tmpbuf);
						} else if (!strcasecmp(comm, "USERINFO")
								   || !strcasecmp(comm, "CLIENTINFO")) {
							char *p = newbuf;
							char *q = rest + len + 1;
							char tmpbuf[1024];

							while (*q && *q != '\001')
								*p++ = *q++;
							*p = '\0';
							jstr2sjstr(tmpbuf, newbuf);
							printf("%s \033[31m%s(%s) = %s\033[m\n", getnow(), comm, nick, tmpbuf);
						} else {
							jstr2sjstr(newbuf, rest + 1);
							printf("%s \033[32mNOTICE from %s: %s\n", getnow(), nick, newbuf);
						}
					}
					break;

				case T_PRIVMSG:
					if (*(rest + 1) == '\001') {
						irc_ctcpfrom(sock, nick, rest + 1);
					} else if (!strcmp(chan, irc_nick)) {
						jstr2sjstr(newbuf, rest + 1);
						printf("%s =%s:%s= %s\n", getnow(), chanbuf, nick, newbuf);
					} else {
						jstr2sjstr(newbuf, rest + 1);
						if (irc_keyword && strstr(newbuf, irc_keyword)) {
							if (keyword_bell)
								_iocs_adpcmout(keyword_bell, 0x0403 /* 15.6KHz, stereo */,
											   bell_length);
							else
								putchar('\a');
							if (irc_keyword_bold) {
								printf("%s <%s:%s> \033[37m%s\033[m\n",
									   getnow(), chanbuf, nick, newbuf);
								return TRUE;
							}
						}
						printf("%s <%s:%s> %s\n", getnow(), chanbuf, nick, newbuf);
					}
					break;

				default:
					printf("%s \033[36mUnsupport command %s\033[m", getnow(), comm);
					break;
			}
		}
#ifdef DEBUG
	} else {
		if (debug)
			printf("\033[46mWhat? %s\033[m", buf);
#endif
	}
	fflush(stdout);
	return TRUE;
}
